#include <functional>
#include <type_traits>
#include <iostream>


#ifndef _GUINYOTE_RWPROPERTY_HPP_
#define _GUINYOTE_RWPROPERTY_HPP_

namespace Guinyote {

template <class PropType>
class TROProperty;

template <class PropType>
class TRWProperty;

template <class  Getter, class Setter>
class ROProperty;

template <class Getter, class Setter>
class RWProperty;


namespace detail {

template <class T>
struct IsProperty : public std::false_type {};

template <class PropType>
struct IsProperty<TROProperty<PropType> > : public std::true_type {};

template <class PropType>
struct IsProperty<TRWProperty<PropType> > : public std::true_type {};

template <class Getter, class Setter>
struct IsProperty<RWProperty<Getter, Setter> > : public std::true_type {};

template <class Getter, class Setter>
struct IsProperty<ROProperty<Getter, Setter> > : public std::true_type {};


template <class T>
struct IsConstMemberGetter : public std::false_type {};

template <class Ret, class Object>
struct IsConstMemberGetter<Ret (Object::*)() const> : public std::true_type {};

template <class T>
struct IsConstMemberSetter : public std::false_type {};

template <class ObjectType, class Arg>
struct IsConstMemberSetter<void (ObjectType::*)(Arg) const> : public std::true_type {};


template <bool cond, class T = void>
struct EnableIf { typedef T type; };

template <class T>
struct EnableIf<false, T> {};

template <bool cond, class T = void>
struct DisableIf {};

template <class T>
struct DisableIf<false, T> 
{
    typedef T type;
};

///Deduces the argument 1 for the setter
template <class ObjectType, class Arg>
auto deduceSetterArg1(void (ObjectType::*Setter)(Arg) const,
const ObjectType * arg = 0) -> typename
		EnableIf<IsConstMemberSetter<void (ObjectType::*)(Arg) const>::value,
		decltype(arg) >::type;


template <class ObjectType, class Arg>
auto deduceSetterArg1(void (ObjectType::*Setter)(Arg),
ObjectType * arg = 0) ->
typename EnableIf<!IsConstMemberSetter<void (ObjectType::*)(Arg)>::value,
		decltype(arg) >::type;

////////Deduces setter arg2

template <class ObjectType, class Arg>
auto deduceSetterArg2(void (ObjectType::*Setter)(Arg) const,
const Arg * arg = 0) -> typename
		EnableIf<IsConstMemberSetter<void (ObjectType::*)(Arg) const>::value,
		typename std::remove_pointer<decltype(arg)>::type >::type;


template <class ObjectType, class Arg>
auto deduceSetterArg2(void (ObjectType::*Setter)(Arg),
Arg * arg = 0) ->
typename EnableIf<!IsConstMemberSetter<void (ObjectType::*)(Arg)>::value,
	typename std::remove_pointer<decltype(arg)>::type >::type;



template <class Ret, class ObjectType>
auto deduceGetterArg(Ret (ObjectType::*Getter)() const, const ObjectType * obj = 0) ->
typename EnableIf<IsConstMemberGetter<Ret (ObjectType::*)() const>::value, decltype(obj) >::type;



template <class Ret, class ObjectType>
auto deduceGetterArg(Ret (ObjectType::*Getter)(), ObjectType * obj = 0) ->
typename EnableIf<!IsConstMemberGetter<Ret (ObjectType::*)()>::value, decltype(obj) >::type;



template <class ObjectType, class PropType>
auto deduceGetterRet(PropType (ObjectType::*Getter)() const, const ObjectType * obj = 0) ->
typename EnableIf<IsConstMemberGetter<PropType (ObjectType::*)() const>::value, decltype((obj->*Getter)())>::type;


template <class ObjectType, class PropType>
auto deduceGetterRet(PropType (ObjectType::*Getter)(), ObjectType * obj = 0) ->
typename EnableIf<!IsConstMemberGetter<PropType (ObjectType::*)()>::value, decltype((obj->*Getter)()) >::type;



} //namespace detail


template <class PropType>
class TRWProperty {
    static_assert(!std::is_reference<PropType>::value, 
                "Property types must not be reference types");
    PropType value_;
public:
    typedef typename std::remove_cv<PropType>::type value_type;
    typedef PropType & GetterReturnType;
    typedef const PropType & ConstGetterReturnType;

    /************************Constructors*************************************/
    
    ///Default constructor
    
    TRWProperty() = default;

    
    ///Construct from other property (rvalue)
    template <class OtherProp>
    TRWProperty(OtherProp && other,
    typename detail::EnableIf<detail::IsProperty<OtherProp>::value>::type * = 0) 
    : value_(std::move(other())) {}
    
    ///Construct from other value (rvalue)
    template <class PropVal>
    TRWProperty(PropVal && val,
    typename detail::EnableIf<!detail::IsProperty<PropVal>::value>::type * = 0) 
    : value_(std::move(val)) {}
    
    ///Construct from other value (lvalue)
    template <class PropVal>
    TRWProperty(const PropVal & value,
    typename detail::EnableIf<!detail::IsProperty<PropVal>::value>::type * = 0) 
    : value_(value) {}
    
    ///Construct from other property (lvalue)
    template <class OtherProp>
    TRWProperty(const OtherProp & prop,
    typename detail::EnableIf<detail::IsProperty<OtherProp>::value>::type * = 0) 
     : value_(prop()) {}
    
    
    
    //***************************Assignment operators*****************************/
    
    ///Assign from other property (rvalue)
    template <class OtherProp>
    typename detail::EnableIf<detail::IsProperty<OtherProp>::value, TRWProperty &>::type
    operator=(OtherProp && other) {
      value_ = std::move(other());
      return *this;
    }
    
    ///Assign from other value (rvalue)
    template <class PropValue>
    typename detail::EnableIf<!detail::IsProperty<PropValue>::value, TRWProperty &>::type
    operator=(PropValue && val) {
      value_ = std::move(val);
      return *this;
    }
    
    ///Assign from other property (lvalue)
    template <class OtherProp>
    typename detail::EnableIf<detail::IsProperty<OtherProp>::value, TRWProperty &>::type
    operator=(const OtherProp & other) {
      value_ = other();
      return *this;
    }
    
    ///Assign from other value (lvalue)
    template <class PropValue>
    typename detail::EnableIf<!detail::IsProperty<PropValue>::value, TRWProperty &>::type
    operator=(const PropValue & val) {
      value_ = val;
      return *this;
    }
       
    /***************************Obtain value operators*****************************/
    ConstGetterReturnType operator()() const {
        return value_;
    }

    GetterReturnType operator()() {
        return value_;
    }
    
    /****************************Implicit conversions******************************/
    operator ConstGetterReturnType() const { return value_; }
    operator GetterReturnType() { return value_; }
    
    /******************************Arithmetic Operators**********************************/
    template <class Prop>
    typename detail::EnableIf<detail::IsProperty<Prop>::value, TRWProperty&>::type
    operator+=(const Prop & other)
    {
	(*this)() += other();
	return *this;
    }
    
    template <class PropValue>
    typename detail::EnableIf<!detail::IsProperty<PropValue>::value, TRWProperty&>::type
    operator+=(const PropValue & value)
    {
	(*this)() += value;
	return *this;
    }
    
    
    TRWProperty &
    operator++()
    {
	++((*this)());
	return *this;
    }
    
    TRWProperty & operator++(int) { ((*this)())++; return *this; }
    
    
    TRWProperty &
    operator--()
    {
	--((*this)());
	return *this;
    }
    
    TRWProperty & operator--(int) { ((*this)())--; return *this; }
    
    /*****************************Relational operators**************************************/
    template <class Prop>
    typename detail::EnableIf<detail::IsProperty<Prop>::value, bool>::type
    operator<(const Prop & other) const
    {
	return (*this)() < other();
    }
    
    template <class PropValue>
    typename detail::EnableIf<!detail::IsProperty<PropValue>::value, bool>::type
    operator<(const PropValue & value) const
    {
	return (*this)() < value;
    }
    
    template <class Prop>
    typename detail::EnableIf<detail::IsProperty<Prop>::value, bool>::type
    operator==(const Prop & other) const
    {
	return (*this)() == other();
    }
    
    template <class PropValue>
    typename detail::EnableIf<!detail::IsProperty<PropValue>::value, bool>::type
    operator==(const PropValue & value) const
    {
	return (*this)() == value;
    }
    
    //******************************Swap***************************
    template <class OtherProp>
    void swap(OtherProp & other)
    {  
	TRWProperty tmp(std::move(other));
	other = std::move(*this);
	*this = std::move(tmp);
    } 
};


template <class PropType>
class TROProperty {
  //FIXME: Make private operator= and friend class the class containing the property
    static_assert(!std::is_reference<PropType>::value, 
                "Property types must not be reference types");
    PropType value_;
public:
    typedef typename std::remove_cv<PropType>::type value_type;
    typedef PropType & GetterReturnType;
    typedef const PropType & ConstGetterReturnType;

    /************************Constructors*************************************/
    
    ///Default constructor
    
    TROProperty() = default;

    
    ///Construct from other property (rvalue)
    template <class OtherProp>
    TROProperty(OtherProp && other,
    typename detail::EnableIf<detail::IsProperty<OtherProp>::value>::type * = 0) 
    : value_(std::move(other())) {}
    
    ///Construct from other value (rvalue)
    template <class PropVal>
    TROProperty(PropVal && val,
    typename detail::EnableIf<!detail::IsProperty<PropVal>::value>::type * = 0) 
    : value_(std::move(val)) {}
    
    ///Construct from other value (lvalue)
    template <class PropVal>
    TROProperty(const PropVal & value,
    typename detail::EnableIf<!detail::IsProperty<PropVal>::value>::type * = 0) 
    : value_(value) {}
    
    ///Construct from other property (lvalue)
    template <class OtherProp>
    TROProperty(const OtherProp & prop,
    typename detail::EnableIf<detail::IsProperty<OtherProp>::value>::type * = 0) 
     : value_(prop()) {}
    

    //***************************Assignment operators*****************************/
    
    ///Assign from other property (rvalue)
    template <class OtherProp>
    typename detail::EnableIf<detail::IsProperty<OtherProp>::value, TROProperty &>::type
    operator=(OtherProp && other) {
      value_ = std::move(other());
      return *this;
    }
    
    ///Assign from other value (rvalue)
    template <class PropValue>
    typename detail::EnableIf<!detail::IsProperty<PropValue>::value, TROProperty &>::type
    operator=(PropValue && val) {
      value_ = std::move(val);
      return *this;
    }
    
    ///Assign from other property (lvalue)
    template <class OtherProp>
    typename detail::EnableIf<detail::IsProperty<OtherProp>::value, TROProperty &>::type
    operator=(const OtherProp & other) {
      value_ = other();
      return *this;
    }
    
    ///Assign from other value (lvalue)
    template <class PropValue>
    typename detail::EnableIf<!detail::IsProperty<PropValue>::value, TROProperty &>::type
    operator=(const PropValue & val) {
      value_ = val;
      return *this;
    }
       
    /***************************Obtain value operators*****************************/
    ConstGetterReturnType operator()() const {
        return value_;
    }

    GetterReturnType operator()() {
        return value_;
    }
    
    /****************************Implicit conversions******************************/
    operator ConstGetterReturnType() const { return value_; }
    operator GetterReturnType() { return value_; }
    
    /******************************Arithmetic Operators**********************************/
    template <class Prop>
    typename detail::EnableIf<detail::IsProperty<Prop>::value, TROProperty&>::type
    operator+=(const Prop & other)
    {
	(*this)() += other();
	return *this;
    }
    
    template <class PropValue>
    typename detail::EnableIf<!detail::IsProperty<PropValue>::value, TROProperty&>::type
    operator+=(const PropValue & value)
    {
	(*this)() += value;
	return *this;
    }
    
    
    TROProperty &
    operator++()
    {
	++((*this)());
	return *this;
    }
    
    TROProperty & operator++(int) { ((*this)())++; return *this; }
    
    TROProperty &
    operator--()
    {
	--((*this)());
	return *this;
    }
    
    TROProperty & operator--(int) { ((*this)())--; return *this; }
    
    /*****************************Relational operators**************************************/
    template <class Prop>
    typename detail::EnableIf<detail::IsProperty<Prop>::value, bool>::type
    operator<(const Prop & other) const
    {
	return (*this)() < other();
    }
    
    template <class PropValue>
    typename detail::EnableIf<!detail::IsProperty<PropValue>::value, bool>::type
    operator<(const PropValue & value) const
    {
	return (*this)() < value;
    }
    
    template <class Prop>
    typename detail::EnableIf<detail::IsProperty<Prop>::value, bool>::type
    operator==(const Prop & other) const
    {
	return (*this)() == other();
    }
    
    template <class PropValue>
    typename detail::EnableIf<!detail::IsProperty<PropValue>::value, bool>::type
    operator==(const PropValue & value) const
    {
	return (*this)() == value;
    }
    
    //******************************Swap***************************
    template <class OtherProp>
    void swap(OtherProp & other)
    {  
	TROProperty tmp(std::move(other));
	other = std::move(*this);
	*this = std::move(tmp);
    } 
};



/************************************************Elaborated properties*******************************************/
template <class Getter, class Setter>
class RWProperty {
    
public:
    typedef typename std::remove_cv<typename std::remove_pointer<
    typename Getter::argument_type>::type>::type ObjectType;
private:
    ObjectType * object_;
    Getter getter_;
    Setter setter_;
public:
    typedef decltype(getter_(object_)) GetterReturnType;
    typedef const GetterReturnType ConstGetterReturnType;
    typedef typename std::remove_cv<GetterReturnType>::type value_type;   
        
    ///Constructor
    RWProperty(ObjectType * object, Getter getter, Setter setter) :
    object_(object), getter_(getter), setter_(setter) {}

    /*
    ///Construct from other property (rvalue)
    template <class OtherProp>
    TRWProperty(ObjectType * object, Getter getter, Setter setter, OtherProp && other,
    typename detail::EnableIf<detail::IsProperty<OtherProp>::value>::type * = 0) 
    : value_(std::move(other())) {}
    
    ///Construct from other value (rvalue)
    template <class PropVal>
    TRWProperty(PropVal && val,
    typename detail::EnableIf<!detail::IsProperty<PropVal>::value>::type * = 0) 
    : value_(std::move(val)) {}
    
    ///Construct from other value (lvalue)
    template <class PropVal>
    TRWProperty(const PropVal & value,
    typename detail::EnableIf<!detail::IsProperty<PropVal>::value>::type * = 0) 
    : value_(value) {}
    
    ///Construct from other property (lvalue)
    template <class OtherProp>
    TRWProperty(const OtherProp & prop,
    typename detail::EnableIf<detail::IsProperty<OtherProp>::value>::type * = 0) 
     : value_(prop()) {}
    */

    //***************************Assignment operators*****************************/
    
    ///Assign from other property (rvalue)
    template <class OtherProp>
    typename detail::EnableIf<detail::IsProperty<OtherProp>::value, RWProperty &>::type
    operator=(OtherProp && other) {
      setter_(object_, std::move(other()));
      return *this;
    }
    
    
    ///Assign from other value (rvalue)
    template <class PropValue>
    typename detail::EnableIf<!detail::IsProperty<PropValue>::value, RWProperty &>::type
    operator=(PropValue && val) {
      setter_(object_, std::move(val));
      return *this;
    }
    
    
    ///Assign from other property (lvalue)
    template <class OtherProp>
    typename detail::EnableIf<detail::IsProperty<OtherProp>::value, RWProperty &>::type
    operator=(const OtherProp & other) {
      setter_(object_, other());
      return *this;
    }
    
    ///Assign from other value (lvalue)
    template <class PropValue>
    typename detail::EnableIf<!detail::IsProperty<PropValue>::value, RWProperty &>::type
    operator=(const PropValue & val) {
      setter_(object_, val);
      return *this;
    }
    
    /***************************Obtain value operators*****************************/
    ConstGetterReturnType operator()() const {
        return getter_(object_);
    }

    GetterReturnType operator()() {
        return getter_(object_);
    }
    
    /****************************Implicit conversions******************************/
    operator ConstGetterReturnType() const { return getter_(object_); }
    operator GetterReturnType() { return getter_(object_); }
    
    /******************************Arithmetic Operators**********************************/
    template <class Prop>
    typename detail::EnableIf<detail::IsProperty<Prop>::value, RWProperty&>::type
    operator+=(const Prop & other)
    {
	auto res = (*this)() + other();
	setter_(object_, res);
	return *this;
    }
    
    template <class PropValue>
    typename detail::EnableIf<!detail::IsProperty<PropValue>::value, RWProperty&>::type
    operator+=(const PropValue & value)
    {
	auto res = (*this)() + value;
	setter_(object_, res);
	return *this;
    }
    
    
    RWProperty &
    operator++()
    {
	auto val = (*this)() + 1;
	setter_(object_, val);
	return *this;
    }
    
    RWProperty & operator++(int) { ++(*this); return *this; }
    
    RWProperty &
    operator--()
    {
	auto val = (*this)() - 1;
	setter_(object_, val);
	return *this;
    }
    
    RWProperty & operator--(int) { --(*this); return *this; }
    
    /*****************************Relational operators**************************************/
    template <class Prop>
    typename detail::EnableIf<detail::IsProperty<Prop>::value, bool>::type
    operator<(const Prop & other) const
    {
	return (*this)() < other();
    }
    
    template <class PropValue>
    typename detail::EnableIf<!detail::IsProperty<PropValue>::value, bool>::type
    operator<(const PropValue & value) const
    {
	return (*this)() < value;
    }
    
    template <class Prop>
    typename detail::EnableIf<detail::IsProperty<Prop>::value, bool>::type
    operator==(const Prop & other) const
    {
	return (*this)() == other();
    }
    
    template <class PropValue>
    typename detail::EnableIf<!detail::IsProperty<PropValue>::value, bool>::type
    operator==(const PropValue & value) const
    {
	return (*this)() == value;
    }
    
    //******************************Swap***************************
    template <class OtherProp>
    void swap(OtherProp & other)
    {  
	value_type tmp(std::move(other()));
	other = std::move(*this);
	*this = std::move(tmp);
    } 
    
};

template <class Getter, class Setter>
class ROProperty {
  public:
    typedef typename std::remove_cv<typename std::remove_pointer<
    typename Getter::argument_type>::type>::type ObjectType;
private:
    ObjectType * object_; 
    Getter getter_;
    Setter setter_;
public:
    typedef decltype(getter_(object_)) GetterReturnType;
    typedef const GetterReturnType ConstGetterReturnType;
    typedef typename std::remove_cv<GetterReturnType>::type value_type;   
public:  
    /************************Constructors*************************************/
    
    ///Default constructor
    //ROProperty() = default;

    ROProperty(ObjectType * object, Getter getter, Setter setter = Setter()) :
    object_(object), getter_(getter), setter_(setter) {}
    
    /*
    ///Construct from other property (rvalue)
    template <class OtherProp>
    TRWProperty(OtherProp && other,
    typename detail::EnableIf<detail::IsProperty<OtherProp>::value>::type * = 0) 
    : value_(std::move(other())) {}
    
    ///Construct from other value (rvalue)
    template <class PropVal>
    TRWProperty(PropVal && val,
    typename detail::EnableIf<!detail::IsProperty<PropVal>::value>::type * = 0) 
    : value_(std::move(val)) {}
    
    ///Construct from other value (lvalue)
    template <class PropVal>
    TRWProperty(const PropVal & value,
    typename detail::EnableIf<!detail::IsProperty<PropVal>::value>::type * = 0) 
    : value_(value) {}
    
    ///Construct from other property (lvalue)
    template <class OtherProp>
    TRWProperty(const OtherProp & prop,
    typename detail::EnableIf<detail::IsProperty<OtherProp>::value>::type * = 0) 
     : value_(prop()) {}
    */

    //***************************Assignment operators*****************************/
    
    ///Assign from other property (rvalue)
    template <class OtherProp>
    typename detail::EnableIf<detail::IsProperty<OtherProp>::value, ROProperty &>::type
    operator=(OtherProp && other) {
      setter_(object_, std::move(other()));
      return *this;
    }
    
    
    ///Assign from other value (rvalue)
    template <class PropValue>
    typename detail::EnableIf<!detail::IsProperty<PropValue>::value, ROProperty &>::type
    operator=(PropValue && val) {
      setter_(object_, std::move(val));
      return *this;
    }
    
    
    ///Assign from other property (lvalue)
    template <class OtherProp>
    typename detail::EnableIf<detail::IsProperty<OtherProp>::value, ROProperty &>::type
    operator=(const OtherProp & other) {
      setter_(object_, other());
      return *this;
    }
    
    ///Assign from other value (lvalue)
    template <class PropValue>
    typename detail::EnableIf<!detail::IsProperty<PropValue>::value, ROProperty &>::type
    operator=(const PropValue & val) {
      setter_(object_, val);
      return *this;
    }
    
    
    /***************************Obtain value operators*****************************/
    ConstGetterReturnType operator()() const {
        return getter_(object_);
    }
    
    
    GetterReturnType operator()() {
        return getter_(object_);
    }
    
    /****************************Implicit conversions******************************/
    operator ConstGetterReturnType() const { return getter_(object_); }
    operator GetterReturnType() { return getter_(object_); }
    
    /******************************Arithmetic Operators**********************************/
    template <class Prop>
    typename detail::EnableIf<detail::IsProperty<Prop>::value, ROProperty&>::type
    operator+=(const Prop & other)
    {
	(*this)() += other();
	return *this;
    }
    
    template <class PropValue>
    typename detail::EnableIf<!detail::IsProperty<PropValue>::value, ROProperty&>::type
    operator+=(const PropValue & value)
    {
	(*this)() += value;
	return *this;
    }
    
    
    ROProperty &
    operator++()
    {
	auto val = (*this)() + 1;
	setter_(object_, val);
	return *this;
    }
    
    ROProperty & operator++(int) { ++(*this); return *this; }
    
    ROProperty &
    operator--()
    {
	auto val = (*this)() - 1;
	setter_(object_, val);
	return *this;
    }
    
    ROProperty & operator--(int) { --(*this); return *this; }
    
    /*****************************Relational operators**************************************/
    template <class Prop>
    typename detail::EnableIf<detail::IsProperty<Prop>::value, bool>::type
    operator<(const Prop & other) const
    {
	return (*this)() < other();
    }
    
    template <class PropValue>
    typename detail::EnableIf<!detail::IsProperty<PropValue>::value, bool>::type
    operator<(const PropValue & value) const
    {
	return (*this)() < value;
    }
    
    template <class Prop>
    typename detail::EnableIf<detail::IsProperty<Prop>::value, bool>::type
    operator==(const Prop & other) const
    {
	return (*this)() == other();
    }
    
    template <class PropValue>
    typename detail::EnableIf<!detail::IsProperty<PropValue>::value, bool>::type
    operator==(const PropValue & value) const
    {
	return (*this)() == value;
    }
    
    //******************************Swap***************************
    template <class OtherProp>
    void swap(OtherProp & other)
    {  
	value_type tmp(std::move(other()));
	other = std::move(*this);
	*this = std::move(tmp);
    } 
};


} //namespace Guinyote


#define GUINYOTE_RWPROPERTY(propname) \
		::Guinyote::RWProperty<\
		std::function<decltype(Guinyote::detail::deduceGetterRet<ClassName>(&ClassName::get##propname)) \
		(decltype(Guinyote::detail::deduceGetterArg(&ClassName::get##propname)))>, \
		std::function<void \
		(decltype(Guinyote::detail::deduceSetterArg1(&ClassName::set##propname)), \
		decltype(Guinyote::detail::deduceSetterArg2(&ClassName::set##propname)))> > propname

#define GUINYOTE_ROPROPERTY(propname) \
::Guinyote::ROProperty<\
  std::function<decltype(Guinyote::detail::deduceGetterRet<ClassName>(&ClassName::get##propname)) \
  (decltype(Guinyote::detail::deduceGetterArg(&ClassName::get##propname)))>, \
  std::function<void \
  (decltype(Guinyote::detail::deduceSetterArg1(&ClassName::set##propname)), \
  decltype(Guinyote::detail::deduceSetterArg2(&ClassName::set##propname)))> > propname


#define GUINYOTE_ROPROPERTY_NOSETTER(propname) \
::Guinyote::ROProperty< \
	 std::function<decltype(Guinyote::detail::deduceGetterRet<ClassName>(&ClassName::get##propname)) \
	 (decltype(Guinyote::detail::deduceGetterArg(&ClassName::get##propname)))>, \
	  std::function<void \
			(ClassName *, \
	decltype(Guinyote::detail::deduceGetterRet<ClassName>(&ClassName::get##propname)))> > propname


#define GUINYOTE_PROPERTYINIT(propname) propname(this, &ClassName::get##propname, &ClassName::set##propname)
#define GUINYOTE_PROPERTYINIT_NOSETTER(propname) \
propname(this, &ClassName::get##propname)


#endif


